package de.mlessmann.certassist.keystore;

import de.mlessmann.certassist.except.JavaSecurityException;
import de.mlessmann.certassist.openssl.CertificatePasswordProvider;
import de.mlessmann.certassist.openssl.CertificateUsage;
import de.mlessmann.certassist.openssl.OpenSSLService;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.OpenOption;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.spec.PKCS8EncodedKeySpec;
import java.util.Arrays;
import java.util.Base64;
import java.util.Collection;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicReference;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;

@Slf4j
@RequiredArgsConstructor
public class KeyStoreManager {

    private static final OpenOption[] CREATE_TRUNCATE = {
        StandardOpenOption.CREATE,
        StandardOpenOption.TRUNCATE_EXISTING,
    };
    private final OpenSSLService openSSLService;
    private final CertificatePasswordProvider passwordProvider;
    private final AtomicReference<CertificateFactory> certFactory = new AtomicReference<>();

    public KeystoreUsage createKeyStore(String keyStorePassphrase, CertificateUsage... serverCerts)
        throws JavaSecurityException {
        try {
            Path keystorePath = Files.createTempFile("keystore", ".jks");
            keystorePath.toFile().deleteOnExit();
            log.debug("Creating keyStore at {}", keystorePath);
            KeyStore keystore = KeyStore.getInstance(KeyStore.getDefaultType());
            keystore.load(null, keyStorePassphrase.toCharArray());

            for (CertificateUsage serverCert : serverCerts) {
                loadPrivateKeyIntoStore(keystore, serverCert);
            }

            // Save the keystore
            try (var keystoreOut = Files.newOutputStream(keystorePath, CREATE_TRUNCATE)) {
                keystore.store(keystoreOut, keyStorePassphrase.toCharArray());
            }
            return new KeystoreResult(keystorePath);
        } catch (IOException | KeyStoreException | NoSuchAlgorithmException | CertificateException e) {
            throw new JavaSecurityException("Failed to create keystore!", e);
        }
    }

    @SneakyThrows
    public KeystoreUsage createTruststore(String truststorePassphrase, CertificateUsage... trustedCertificates) {
        return createTruststore(truststorePassphrase, Arrays.asList(trustedCertificates));
    }

    public KeystoreUsage createTruststore(
        String truststorePassphrase,
        Collection<CertificateUsage> trustedCertificates
    ) {
        try {
            Path truststorePath = Files.createTempFile("truststore", ".jks");
            truststorePath.toFile().deleteOnExit();
            log.debug("Creating truststore at {}", truststorePath);
            KeyStore truststore = KeyStore.getInstance(KeyStore.getDefaultType());
            truststore.load(null, truststorePassphrase.toCharArray());

            for (CertificateUsage serverCert : trustedCertificates) {
                loadPublicKeyIntoTrust(truststore, serverCert);
            }

            try (var outputStream = Files.newOutputStream(truststorePath, CREATE_TRUNCATE)) {
                truststore.store(outputStream, truststorePassphrase.toCharArray());
            }
            return new KeystoreResult(truststorePath);
        } catch (IOException | KeyStoreException | CertificateException | NoSuchAlgorithmException e) {
            throw new IllegalStateException("Failed to create truststore!", e);
        }
    }

    public void loadPrivateKeyIntoStore(KeyStore keyStore, CertificateUsage cert) throws KeyStoreException {
        String passphrase = passwordProvider.getPasswordFor(cert.fingerprint());
        var pKey = loadPrivateKey(
            Objects.requireNonNull(cert.certificateKeyPath(), "Certificate does not have a key to import."),
            passphrase
        );
        var certChain = loadCertificateChain(cert.fullchainPath());
        keyStore.setKeyEntry("KEY;%s".formatted(cert.fingerprint()), pKey, null, certChain);
    }

    public void loadPublicKeyIntoTrust(KeyStore keyStore, CertificateUsage cert) throws KeyStoreException {
        try (var inputStream = Files.newInputStream(cert.certificatePath())) {
            Collection<? extends Certificate> parsedCerts = getX509Factory().generateCertificates(inputStream);
            if (parsedCerts.size() != 1) {
                throw new IllegalStateException(
                    "CertificateUsage parsed to not exactly one result! " + parsedCerts.size()
                );
            }
            Certificate parsedCert = parsedCerts.stream().toList().getFirst();
            if (!(parsedCert instanceof X509Certificate x509cert)) {
                throw new IllegalStateException("CertificateUsage did not parse to X509 format??");
            }
            keyStore.setCertificateEntry("TRUST;%s".formatted(cert.fingerprint()), x509cert);
        } catch (IOException e) {
            throw new IllegalStateException("CertificateUsage could not be read??", e);
        } catch (CertificateException e) {
            throw new IllegalStateException("CertificateUsage is not a parsable certificate??");
        }
    }

    @SneakyThrows
    private PrivateKey loadPrivateKey(Path privateKey, String passphrase) {
        String pemContent;
        if (openSSLService.isKeyEncrypted(privateKey)) {
            pemContent = openSSLService.readDecryptedKey(privateKey, passphrase);
        } else {
            pemContent = Files.readString(privateKey);
        }

        String privateKeyPEM = pemContent
            .replaceAll(".*?-----BEGIN PRIVATE KEY-----", "")
            .replace("-----END PRIVATE KEY-----", "")
            .replaceAll("\\s", "");
        byte[] decodedKey = Base64.getDecoder().decode(privateKeyPEM);
        return KeyFactory.getInstance("RSA").generatePrivate(new PKCS8EncodedKeySpec(decodedKey));
    }

    @SneakyThrows
    private Certificate[] loadCertificateChain(Path certChainPath) {
        // Load the certificate chain from a PEM file
        try (var fis = Files.newInputStream(certChainPath)) {
            CertificateFactory certFactory = CertificateFactory.getInstance("X.509");
            return certFactory.generateCertificates(fis).toArray(Certificate[]::new);
        }
    }

    protected final synchronized CertificateFactory getX509Factory() {
        CertificateFactory instance = certFactory.get();
        if (instance == null) {
            try {
                instance = CertificateFactory.getInstance("X.509");
                certFactory.set(instance);
            } catch (CertificateException e) {
                throw new IllegalStateException("X.509 factory missing??", e);
            }
        }
        return instance;
    }
}
