package de.mlessmann.certassist;

import static org.assertj.core.api.Assertions.assertThat;

import de.mlessmann.certassist.keystore.KeyStoreManager;
import de.mlessmann.certassist.models.Certificate;
import de.mlessmann.certassist.models.CertificateInfo;
import de.mlessmann.certassist.models.CertificateInfoExtension;
import de.mlessmann.certassist.models.CertificateInfoSubject;
import de.mlessmann.certassist.openssl.CertificateProvider;
import de.mlessmann.certassist.openssl.CertificateUsage;
import de.mlessmann.certassist.repositories.CertificateRepository;
import de.mlessmann.certassist.service.CertificateCreationService;
import java.nio.file.Path;
import java.util.List;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

@SpringBootTest
public class CertificateServiceTest {

    public static final Path LETSENCRYPT_CERT_PATH = Path.of("src/test/resources/letsencrypt");

    @Autowired
    CertificateRepository certificateRepository;

    @Autowired
    CertificateCreationService certificateService;

    @Autowired
    KeyStoreManager keyStoreManager;

    @Autowired
    CertificateProvider certificateProvider;

    @BeforeEach
    void setUp() {
        certificateRepository.deleteAll();
    }

    @Test
    void testCanCreateCertificate() {
        var request = CertificateInfo
            .builder()
            .type(CertificateInfo.RequestType.STANDALONE_CERTIFICATE)
            .subject(
                CertificateInfoSubject
                    .builder()
                    .commonName("cert.creation")
                    .country("DE")
                    .state("SH")
                    .locality("HH")
                    .organization("Crazy-Cats")
            )
            .extension(CertificateInfoExtension.builder().alternativeDnsNames("test2.home", "test3.home"))
            .build();
        var cert = certificateService.createCertificate(request);
        assertThat(cert).isNotNull();
        assertThat(cert.getId()).isGreaterThan("0");
    }

    @Test
    void testCanImportCertificate() {
        Path certDir = TestOpenSSLService.TEST_CERT_PATH;
        var importedCert = certificateService.importCertificate(
            certDir.resolve("x509forImport.pem"),
            certDir.resolve("x509forImport.key.pem"),
            TestOpenSSLService.TEST_CERT_PASSPHRASE
        );
        assertThat(importedCert).isNotNull();
        assertThat(importedCert.getId()).isGreaterThan("0");
    }

    @Test
    void testCanImportTrustCertificate() {
        Path cert = LETSENCRYPT_CERT_PATH.resolve("2024-11-isrgrootx1.pem");
        var importedCert = certificateService.importCertificate(cert, null, null);
        assertThat(importedCert).isNotNull();
        assertThat(importedCert.getId()).isGreaterThan("0");
    }

    @Test
    void testCanImportTrustBundles() throws Exception {
        Path cert = LETSENCRYPT_CERT_PATH.resolve("2024-11-letsencrypt.bundle.pem");
        List<Certificate> importedCerts = certificateService.importCertificateTrustBundle(cert);
        List<CertificateUsage> tmpUsableCerts = importedCerts
            .stream()
            .map(Certificate::getFingerprint)
            .map(certificateProvider::requestCertificateUsage)
            .toList();

        try (var store = keyStoreManager.createTruststore("changeit", tmpUsableCerts)) {
            assertThat(store).isNotNull();
            assertThat(store.truststorePath()).isRegularFile();
        } finally {
            tmpUsableCerts.forEach(CertificateUsage::close);
        }
    }
}
