package de.mlessmann.certassist.repositories;

import static org.assertj.core.api.Assertions.assertThat;

import de.mlessmann.certassist.models.Certificate;
import de.mlessmann.certassist.models.CertificateExtension;
import de.mlessmann.certassist.models.CertificateType;
import jakarta.transaction.Transactional;

import java.time.OffsetDateTime;
import java.util.List;
import java.util.stream.StreamSupport;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

@SpringBootTest
class CertificateRepositoryTest {

    @Autowired
    private CertificateRepository repository;

    @Autowired
    private CertificateExtensionRepository extensionRepository;

    @BeforeEach
    void cleanUp() {
        extensionRepository.deleteAll();
        repository.deleteAll();
    }

    private Certificate getCertificate() {
        final Certificate certificate = new Certificate();
        certificate.setSubjectCommonName("test-cn");
        certificate.setType(CertificateType.SIGNED_CERT);
        certificate.setRequestedKeyLength(1);
        certificate.setNotBefore(OffsetDateTime.now());
        certificate.setNotAfter(OffsetDateTime.now().plusDays(1));

        return certificate;
    }

    @Test
    @Transactional
    void findCertificate() {
        final Certificate certificate = getCertificate();
        repository.save(certificate);

        Certificate foundCertificate = repository.findById(certificate.getId()).orElseThrow();
        assertThat(foundCertificate.getSubjectCommonName()).isEqualTo("test-cn");
        assertThat(foundCertificate.getType()).isEqualTo(CertificateType.SIGNED_CERT);
    }

    @Test
    @Transactional
    void findCertificateWithExtension() {
        final Certificate certificate = getCertificate();
        certificate.setFingerprint("test-fingerprint");
        CertificateExtension extension = new CertificateExtension();
        extension.setValue("test-ext-value");
        certificate.setCertificateExtension(List.of(extension));
        repository.save(certificate);

        assertThat(repository.findById(certificate.getId()).orElseThrow().getCertificateExtension()).hasSize(1);
        List<CertificateExtension> extensions = StreamSupport
            .stream(extensionRepository.findAll().spliterator(), false)
            .toList();
        assertThat(extensions).hasSize(1);
        assertThat(extensions).singleElement().satisfies(ce -> assertThat(ce.getValue()).isEqualTo("test-ext-value"));
    }
}
